#define _UCRT
#undef _WIN32_WINNT
#define _WIN32_WINNT 0x501
#include <errno.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <wchar.h>
#include <windows.h>

// Defining this macro will only affect the GCC compiler and can reduce the size of the utility. As soon as you define it you have to add these linker options:
//   -Wl,-e,__start -nostartfiles
#define USE_START

#if defined(__GNUC__) && defined(USE_START)
# define QUIT(_status) exit(_status)
# ifdef __i386
void _start(void);
void _start(void)
# else
void __start(void);
void __start(void)
# endif
#else
# define QUIT(_status) return (_status)
int main(void)
#endif
{
  int argc;
  wchar_t **argv = CommandLineToArgvW(GetCommandLineW(), &argc);
  if (!argv)
    QUIT(1);

  if (argc < 2 || wcscmp(argv[1], L"/?") == 0)
  {
    LocalFree(argv);
    puts("Create a HEX dump of a sector of a drive. Each hexadecimal expression of a\n"
         " byte consists of two digits and is written to a new line.\n"
         " NOTE: Certain drives require the tool to be executed as Administrator.\n\n"
         "SYNTAX\n"
         " drivedump \"drive\" [sector_index]\n"
         "   drive          Logical drive (such as \"C:\"), or physical disk drive (such\n"
         "                   as \"PhysicalDrive0\").\n"
         "   sector_index   (optional) Zero-based index of the sector to be read.\n"
         "                   Default index is 0.");
    QUIT(1);
  }

  long long idx = 0;
  if (argc > 2)
  {
    errno = 0;
    wchar_t *endptr;
    idx = wcstoll(argv[2], &endptr, 0);
    if (errno || *endptr)
    {
      LocalFree(argv);
      QUIT(1);
    }
  }

  wchar_t name[24] = L"\\\\.\\";
  wcsncpy(name + 4, argv[1], 16);
  LocalFree(argv);
  HANDLE drive = CreateFileW(name, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, NULL, OPEN_EXISTING, 0, NULL);
  if(drive == INVALID_HANDLE_VALUE)
    QUIT(1);

  DISK_GEOMETRY geo;
  DWORD cb;
  BYTE *sector;
  if (!DeviceIoControl(drive, IOCTL_DISK_GET_DRIVE_GEOMETRY, NULL, 0, &geo, sizeof(geo), &cb, NULL) ||
      !SetFilePointerEx(drive, (LARGE_INTEGER){ .QuadPart = idx * geo.BytesPerSector }, NULL, FILE_BEGIN) ||
      !(sector = (BYTE*)malloc(geo.BytesPerSector)))
  {
    CloseHandle(drive);
    QUIT(1);
  }

  const int r = ReadFile(drive, sector, geo.BytesPerSector, &cb, NULL);
  CloseHandle(drive);
  if (r)
    for (const BYTE *iter = sector, *const end = sector + geo.BytesPerSector; iter < end; printf("%02X\n", *iter++));

  free(sector);
  QUIT(!r);
}
